#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>

#include "fskeycode.h"
#include "fswin32keymap.h"


static FsWin32KeyMapper fsKeyMapper;


HWND hWnd=NULL;
HGLRC hRC;
HPALETTE plt;
HDC wDC;





static LONG WINAPI WindowFunc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp);
static void SetPixelFormat(HDC dc,int useDoubleBuffer);
static HPALETTE YsCreatePalette(HDC dc);
static void InitializeOpenGL(HWND wnd);



// For OpenGL set up >>
static int doubleBuffer=0;
// For OpenGL set up <<


#define NKEYBUF 256
static int keyBuffer[NKEYBUF];
static int nKeyBufUsed=0;
static int charBuffer[NKEYBUF];
static int nCharBufUsed=0;


#define WINSTYLE WS_OVERLAPPED|WS_CAPTION|WS_VISIBLE|WS_SYSMENU|WS_THICKFRAME|WS_MINIMIZEBOX|WS_MAXIMIZEBOX
#define WINCLASS "24-780"
#define WINNAME ""

void FsOpenWindow(int wid,int hei,int useDoubleBuffer)
{
	WNDCLASS wc;
	HINSTANCE inst;

	inst=GetModuleHandle(NULL);

	wc.style=CS_OWNDC|CS_BYTEALIGNWINDOW;
	wc.lpfnWndProc=WindowFunc;
	wc.cbClsExtra=0;
	wc.cbWndExtra=0;
	wc.hInstance=(HINSTANCE)inst;
	wc.hIcon=LoadIcon(inst,"MAINICON");
	wc.hCursor=LoadCursor(NULL,IDC_ARROW);
	wc.hbrBackground=(HBRUSH)GetStockObject(WHITE_BRUSH);
	wc.lpszMenuName=NULL;
	wc.lpszClassName=WINCLASS;
	if(RegisterClass(&wc))
	{
		doubleBuffer=useDoubleBuffer;


		RECT rc;
		rc.left  =64;
		rc.top   =64;
		rc.right =(unsigned long)(64+wid-1);
		rc.bottom=(unsigned long)(64+hei-1);
		AdjustWindowRect(&rc,WINSTYLE,FALSE);
		wid  =rc.right-rc.left+1;
		hei  =rc.bottom-rc.top+1;

		hWnd=CreateWindow(WINCLASS,WINNAME,WINSTYLE,0,0,wid,hei,NULL,NULL,inst,NULL);
		if(hWnd!=NULL)
		{
			InitializeOpenGL(hWnd);

			ShowWindow(hWnd,SW_SHOWNORMAL);
			UpdateWindow(hWnd);
		}
		else
		{
			printf("Could not open window.\n");
			exit(1);
		}
	}
}

void FsGetWindowSize(int &wid,int &hei)
{
	RECT rect;
	GetClientRect(hWnd,&rect);
	wid=rect.right;
	hei=rect.bottom;
}

void FsPollDevice(void)
{
	MSG msg;
	while(PeekMessage(&msg,hWnd,0,0,PM_REMOVE))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
}

void FsSleep(int ms)
{
	if(ms>0)
	{
		Sleep(ms);
	}
}

int FsPassedTime(void)
{
	static unsigned int lastTick;
	unsigned int tick,passed;

	tick=GetTickCount();
	if(lastTick<tick)
	{
		passed=tick-lastTick;
	}
	else
	{
		passed=1;
	}
	lastTick=tick;

	return passed;
}

int FsInkey(void)
{
	if(nKeyBufUsed>0)
	{
		int i,keyCode;
		keyCode=keyBuffer[0];
		nKeyBufUsed--;
		for(i=0; i<nKeyBufUsed; i++)
		{
			keyBuffer[i]=keyBuffer[i+1];
		}
		return keyCode;
	}
	return 0;
}

int FsInkeyChar(void)
{
	if(nCharBufUsed>0)
	{
		int i,asciiCode;
		asciiCode=charBuffer[0];
		nCharBufUsed--;
		for(i=0; i<nCharBufUsed; i++)
		{
			charBuffer[i]=charBuffer[i+1];
		}
		return asciiCode;
	}
	return 0;
}

int FsGetKeyState(int fsKeyCode)
{
	int vk;
	vk=fsKeyMapper.FsKeyToVk(fsKeyCode);
	return (GetKeyState(vk)&0x8000)!=0;
}

void FsGetMouseState(int &lb,int &mb,int &rb,int &mx,int &my)
{
	POINT cur;
	GetCursorPos(&cur);
	ScreenToClient(hWnd,&cur);

	mx=cur.x;
	my=cur.y;

	lb=((GetKeyState(VK_LBUTTON)&0x8000)!=0 ? 1 : 0);
	mb=((GetKeyState(VK_MBUTTON)&0x8000)!=0 ? 1 : 0);
	rb=((GetKeyState(VK_RBUTTON)&0x8000)!=0 ? 1 : 0);
}

void FsSwapBuffers(void)
{
	HDC dc;
	glFlush();
	dc=wglGetCurrentDC();
	SwapBuffers(dc);
}

static LONG WINAPI WindowFunc(HWND wnd,UINT msg,WPARAM wp,LPARAM lp)
{
	switch(msg)
	{
	case WM_QUERYNEWPALETTE:
	case WM_PALETTECHANGED:
		if(plt!=NULL)
		{
			SelectPalette(GetDC(wnd),plt,FALSE);
			RealizePalette(GetDC(wnd));
		}
		return DefWindowProc(wnd,msg,wp,lp);
	case WM_CREATE:
		wDC=GetDC(wnd);
		SetPixelFormat(wDC,doubleBuffer);
		hRC=wglCreateContext(wDC);
		wglMakeCurrent(wDC,hRC);
		InitializeOpenGL(wnd);
		break;
	case WM_SIZE:
		wDC=GetDC(wnd);
		wglMakeCurrent(wDC,hRC);
		break;
	case WM_PAINT:
		wDC=GetDC(wnd);
		wglMakeCurrent(wDC,hRC);
		return DefWindowProc(wnd,msg,wp,lp);
	case WM_COMMAND:
		break;
	case WM_DESTROY:
		exit(1);
		break;
	case WM_KEYDOWN:
		if(nKeyBufUsed<NKEYBUF)
		{
			int keyCode;
			keyCode=fsKeyMapper.VkToFsKey(wp);
			if(keyCode!=0)
			{
				keyBuffer[nKeyBufUsed++]=keyCode;
			}
		}
		break;
	case WM_CHAR:
		if(nCharBufUsed<NKEYBUF)
		{
			charBuffer[nCharBufUsed++]=wp;
		}
		break;
	case WM_ERASEBKGND:
		return 1;

	default:
		return DefWindowProc(wnd,msg,wp,lp);
	}
	return 1;
}

static void SetPixelFormat(HDC dc,int useDoubleBuffer)
{
	static PIXELFORMATDESCRIPTOR pfd=
	{
		sizeof(PIXELFORMATDESCRIPTOR),
		1,
		PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL,
		PFD_TYPE_RGBA,
		24,
		0,0,0,0,0,0,
		0,
		0,
		0,
		0,0,0,0,
		32,
		0,
		0,
		PFD_MAIN_PLANE,
		0,
		0,0,0
	};

	if(useDoubleBuffer!=0)
	{
		pfd.dwFlags|=PFD_DOUBLEBUFFER;
	}

	int pfm;
	pfm=ChoosePixelFormat(dc,&pfd);
	if(pfm!=0)
	{
		if(SetPixelFormat(dc,pfm,&pfd)!=FALSE)
		{
			plt=YsCreatePalette(dc);
			SelectPalette(dc,plt,FALSE);
			RealizePalette(dc);
			return;
		}
	}
	MessageBox(NULL,"Error In YsSetPixelFormat.",NULL,MB_OK);
	exit(1);
}

static unsigned char YsPalVal(unsigned long n,unsigned bit,unsigned sft)
{
	unsigned long msk;
	n>>=sft;
	msk=(1<<bit)-1;
	n&=msk;
	return (unsigned char)(n*255/msk);
}

/* ? lp=LocalAlloc(LMEM_FIXED,sizeof(LOGPALETTE)+n*sizeof(PALETTEENTRY)); */
/* ? LocalFree(lp); */

static HPALETTE YsCreatePalette(HDC dc)
{
	HPALETTE neo;
    PIXELFORMATDESCRIPTOR pfd;
    LOGPALETTE *lp;
    int n,i;

    n=GetPixelFormat(dc);
    DescribePixelFormat(dc,n,sizeof(PIXELFORMATDESCRIPTOR),&pfd);

    if(pfd.dwFlags & PFD_NEED_PALETTE)
    {
        n=1<<pfd.cColorBits;
        lp=(LOGPALETTE *)malloc(sizeof(LOGPALETTE)+n*sizeof(PALETTEENTRY));
        lp->palVersion=0x300;
        lp->palNumEntries=n;
        for (i=0; i<n; i++)
        {
            lp->palPalEntry[i].peRed  =YsPalVal(i,pfd.cRedBits,pfd.cRedShift);
            lp->palPalEntry[i].peGreen=YsPalVal(i,pfd.cGreenBits,pfd.cGreenShift);
            lp->palPalEntry[i].peBlue =YsPalVal(i,pfd.cBlueBits,pfd.cBlueShift);
            lp->palPalEntry[i].peFlags=0;
        }

        neo=CreatePalette(lp);
		free(lp);
	    return neo;
    }
	return NULL;
}

static void InitializeOpenGL(HWND wnd)
{
	RECT rect;
	float aspect;

    GetClientRect(wnd,&rect);

    glClearColor(1.0F,1.0F,1.0F,1.0F);
    glClearDepth(1.0F);

	glViewport(0,0,rect.right,rect.bottom);

	glDisable(GL_DEPTH_TEST);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

    glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
    aspect=(float)rect.right/rect.bottom;;
    /* gluPerspective(45.0F,aspect,0.5,17.0F); */
	/* glFrustum(-20.0F,20.0F,-20.0F,20.0F,  50,200.0F); Perspective̍ؔ */

	glOrtho(0,(float)rect.right-1,(float)rect.bottom-1,0,-1,1);

	glScalef(1.0F,1.0F,1.0F);
	glShadeModel(GL_FLAT);

	glPointSize(1);

	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	glColor3ub(0,0,0);
}
